"""
totalcloud.py - Contains the functions to upload supported GAV API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList

BASE_AWS_FIELDS = {
    "resourceId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    "connectorUuids": types.String().with_variant(
        TEXT(charset="utf8"), "mysql", "mariadb"
    ),  # BaseList
    "created": types.DateTime(),
    "cloudAccountId": types.BigInteger(),
    "additionalDetails": types.String().with_variant(
        TEXT(charset="utf8"), "mysql", "mariadb"
    ),
    "uuid": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    "connectorUuid": types.String().with_variant(
        TEXT(charset="utf8"), "mysql", "mariadb"
    ),
    "createdOn": types.DateTime(),
    "tags": types.String().with_variant(
        TEXT(charset="utf8"), "mysql", "mariadb"
    ),  # BaseList
    "remediationEnabled": types.Boolean(),
    "lastUpdated": types.DateTime(),
    "cloudType": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    "region": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    "resourceType": types.String().with_variant(
        TEXT(charset="utf8"), "mysql", "mariadb"
    ),
    "accountAlias": types.String().with_variant(
        TEXT(charset="utf8"), "mysql", "mariadb"
    ),
    "controlsFailed": types.Integer(),
    "qualysTags": types.String().with_variant(
        TEXT(charset="utf8"), "mysql", "mariadb"
    ),  # BaseList
}


def upload_totalcloud_aws_connectors(
    connectors: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_connectors",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_connectors(provider='aws') to SQL.

    Args:
        connectors (BaseList): The BaseList of Connectors to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "description": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "provider": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "state": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "totalAssets": types.Integer(),
        "lastSyncedOn": types.DateTime(),
        "nextSyncedOn": types.DateTime(),
        "remediationEnabled": types.Boolean(),
        "qualysTags": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isGovCloud": types.Boolean(),
        "isChinaRegion": types.Boolean(),
        "awsAccountId": types.BigInteger(),
        "accountAlias": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isDisabled": types.Boolean(),
        "pollingFrequency": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "error": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "baseAccountId": types.BigInteger(),
        "externalId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "arn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "portalConnectorUuid": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isPortalConnector": types.Boolean(),
        "groups": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(connector) for connector in connectors])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_azure_connectors(
    connectors: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_azure_connectors",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_connectors(provider='azure') to SQL.

    Args:
        connectors (BaseList): The BaseList of Connectors to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "description": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "provider": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "state": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "totalAssets": types.Integer(),
        "lastSyncedOn": types.DateTime(),
        "nextSyncedOn": types.DateTime(),
        "remediationEnabled": types.Boolean(),
        "qualysTags": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isGovCloud": types.Boolean(),
        "isChinaRegion": types.Boolean(),
        "accountAlias": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isDisabled": types.Boolean(),
        "pollingFrequency": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "error": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "externalId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "portalConnectorUuid": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isPortalConnector": types.Boolean(),
        "subscriptionId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "tenantId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "applicationId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subscriptionName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "directoryId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "groups": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(connector) for connector in connectors])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_gcp_connectors(
    connectors: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_gcp_connectors",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_connectors(provider='gcp') to SQL.

    Args:
        connectors (BaseList): The BaseList of Connectors to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "description": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "provider": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "state": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "totalAssets": types.Integer(),
        "lastSyncedOn": types.DateTime(),
        "nextSyncedOn": types.DateTime(),
        "remediationEnabled": types.Boolean(),
        "qualysTags": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isGovCloud": types.Boolean(),
        "isChinaRegion": types.Boolean(),
        "accountAlias": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isDisabled": types.Boolean(),
        "pollingFrequency": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "error": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "externalId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "portalConnectorUuid": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isPortalConnector": types.Boolean(),
        "groups": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "projectId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(connector) for connector in connectors])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_control_metadata(
    controls: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_control_metadata",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_metadata() to SQL.

    Args:
        controls (BaseList): The BaseList of Controls to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "cid": types.BigInteger(),
        "controlName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "created": types.DateTime(),
        "modified": types.DateTime(),
        "controlType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "provider": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isCustomizable": types.Boolean(),
        "serviceType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "criticality": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "evaluation_description": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "evaluation_passMessage": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "evaluation_failMessage": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "evaluation_criteria": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "specification": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rationale": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "manualRemediation": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "references": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "buildTimeRemediation": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "resourceType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "remediationEnabled": types.Boolean(),
        "policyNames": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "executionType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "workflowBased": types.Boolean(),
        "templateType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(control) for control in controls])

    # Drop the evaluation column:
    df.drop(columns=["evaluation"], inplace=True)

    # Upload the data:

    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_ec2(
    ec2s: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_ec2_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='VM') to SQL.

    Args:
        ec2s (BaseList): The BaseList of EC2s to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "subnetId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "arsScore": types.Integer(),
        "availabilityZone": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "instanceId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "instanceState": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "networkInterfaceAddresses": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "vulnerabilityStats_severity_1": types.Integer(),
        "vulnerabilityStats_severity_2": types.Integer(),
        "vulnerabilityStats_severity_3": types.Integer(),
        "vulnerabilityStats_severity_4": types.Integer(),
        "vulnerabilityStats_severity_5": types.Integer(),
        "vulnerabilityStats_typeDetected_Confirmed": types.Integer(),
        "vulnerabilityStats_typeDetected_Potential": types.Integer(),
        "vulnerabilityStats_typeDetected_Informational": types.Integer(),
        "vulnerabilityStats_typeDetected_totalVulnerability": types.Integer(),
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "events": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "iamInstanceProfileRoleDetails_profileName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "iamInstanceProfileRoleDetails_profileArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "iamInstanceProfileRoleDetails_roleArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "iamInstanceProfileRoleDetails_roleName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "imageId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "publicIpAddress": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "instanceType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "ipAddress": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "publicDnsName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "instanceStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "iamInstanceProfile_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "iamInstanceProfile_id": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "iamInstanceProfile_arn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "privateIpAddress": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "launchTime": types.DateTime(),
        "classifications": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "spotInstanceRequestId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "elasticIpAddress": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "secondaryPrivateIpAddress": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "securityGroups": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "privateDnsName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "criticalityScore": types.Integer(),
        "vulnerabilities": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(ec2) for ec2 in ec2s])

    # Drop vulnerabilityStats, iamInstanceProfileRoleDetails, iamInstanceProfile columns:
    df.drop(
        columns=[
            "vulnerabilityStats",
            "iamInstanceProfileRoleDetails",
            "iamInstanceProfile",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_s3(
    buckets: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_s3_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='S3') to SQL.

    Args:
        buckets (BaseList): The BaseList of S3 Buckets to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "bucketName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "bucketCreationDateStr": types.DateTime(),
        "s3GrantList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "ownerName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "bucketPolicy": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "bucketOwnerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(bucket) for bucket in buckets])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_acl(
    acls: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_acl_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='acl') to SQL.

    Args:
        acls (BaseList): The BaseList of ACLs to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "associations": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "ipPermissionEgressList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "ipPermissionList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "isDefault": types.Boolean(),
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "networkAclId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(acl) for acl in acls])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_rds(
    rds: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_rds_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='rds') to SQL.

    Args:
        rds (BaseList): The BaseList of RDSs to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "subnetId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subnetGroup_dbSubnetGroupStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subnetGroup_dbSubnetGroupName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subnetGroup_dbSubnetVpcId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subnetGroup_subnetList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "subnetGroup_dbsubnetGroupArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dbName": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "availabilityZone": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "securityGroupId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dbClusterIdentifier": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "caCertificateIdentifier": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "engineVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "masterUsername": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dbInstanceIdentifier": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dbSecurityGroupList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "backupRetentionPeriod": types.Integer(),
        "kmsKeyId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "latestRestorableTime": types.DateTime(),
        "dbInstancePort": types.Integer(),
        "enhancedMonitoringEnabled": types.Boolean(),
        "dbiResourceId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "preferredBackupWindow": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "endpoint_hostedZoneId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "endpoint_address": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "endpoint_port": types.Integer(),
        "engine": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "publiclyAccessible": types.Boolean(),
        "arn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "instanceCreatedTime": types.DateTime(),
        "multiAZ": types.Boolean(),
        "instanceClass": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "encrypted": types.Boolean(),
        "iamDatabaseAuthenticationEnabled": types.Boolean(),
        "licenseModel": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "preferredMaintenanceWindow": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "storageType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(rd) for rd in rds])

    # Drop the subnetGroup, endpoint columns:
    df.drop(columns=["subnetGroup", "endpoint"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_iamuser(
    iam_users: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_iamuser_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='iamuser') to SQL.

    Args:
        iam_users (BaseList): The BaseList of IAM Users to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "userPolicies": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "userAttachedPolicies": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "userDto_accessKey1Active": types.Boolean(),
        "userDto_accessKey2Active": types.Boolean(),
        "userDto_cert1LastRotated": types.DateTime(),
        "userDto_cert2LastRotated": types.DateTime(),
        "userDto_cert1Active": types.Boolean(),
        "userDto_cert2Active": types.Boolean(),
        "userDto_accessKey1LastRotated": types.DateTime(),
        "userDto_accessKey2LastRotated": types.DateTime(),
        "userDto_accesKey1LastUsedService": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userDto_accessKey2LastUsedService": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userDto_path": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userDto_accessKeys": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "userDto_passwordLastUsed": types.DateTime(),
        "userDto_passwordNextRotation": types.DateTime(),
        "userDto_accessKey1LastUsedRegion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userDto_accessKey2LastUsedRegion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userDto_arn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userDto_userCreationTime": types.DateTime(),
        "userDto_mfaActive": types.Boolean(),
        "userDto_passwordEnabled": types.Boolean(),
        "userDto_accessKey1LastUsed": types.DateTime(),
        "userDto_accessKey2LastUsed": types.DateTime(),
        "userDto_userId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userDto_createdDate": types.DateTime(),
        "userDto_passwordLastChanged": types.DateTime(),
        "userDto_username": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "userGroups": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "classifications": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "userInlinePolicies": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "user": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(iam_user) for iam_user in iam_users])

    # Drop the userDto column:
    df.drop(columns=["userDto"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_vpc(
    vpcs: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_vpc_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='vpc') to SQL.

    Args:
        vpcs (BaseList): The BaseList of VPCs to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "instanceTenancy": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "cidrBlock": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isDefault": types.Boolean(),
        "ipv6CidrBlockAssociationSet": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(vpc) for vpc in vpcs])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_securitygroup(
    security_groups: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_securitygroup_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='sg') to SQL.

    Args:
        security_groups (BaseList): The BaseList of Security Groups to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "groupId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "description": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "vulnerable": types.Boolean(),
        "ipPermissionList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "ipPermissionEgressList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "groupName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [prepare_dataclass(security_group) for security_group in security_groups]
    )

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_lambda(
    lambdas: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_lambda_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='lambda') to SQL.

    Args:
        lambdas (BaseList): The BaseList of Lambdas to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "functionName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "associations": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "handler": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rawFunctionPolicy": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "masterArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "aliases": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "assumeRolePolicyDocument": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "kmsKeyArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "tracingConfig": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "description": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "inlinePolicies": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "codeSize": types.Integer(),
        "timeout": types.Integer(),
        "codeSha256": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "roleArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "layers": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "functionArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "functionVersions": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "runtime": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "functionUrlConfigs": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "revisionId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "classifications": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "managedPolicies": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "memorySize": types.Integer(),
        "eventSourceMappings": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "roleName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "lastModified": types.DateTime(),
        "version": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(lambda_) for lambda_ in lambdas])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_subnet(
    subnets: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_subnet_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='subnet') to SQL.

    Args:
        subnets (BaseList): The BaseList of Subnets to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "subnetId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "availabilityZone": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "assignIpv6AddressOnCreation": types.Boolean(),
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "ipv6CidrBlockAssociationSet": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "cidrBlock": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "availableIpAddressCount": types.Integer(),
        "mapPublicIpOnLaunch": types.Boolean(),
        "defaultForAz": types.Boolean(),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(subnet) for subnet in subnets])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_internetgateway(
    internet_gateways: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_internetgateway_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='internetgateway') to SQL.

    Args:
        internet_gateways (BaseList): The BaseList of Internet Gateways to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "internetGatewayId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "attachments": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [prepare_dataclass(internet_gateway) for internet_gateway in internet_gateways]
    )

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_loadbalancer(
    load_balancers: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_loadbalancer_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='loadbalancer') to SQL.

    Args:
        load_balancers (BaseList): The BaseList of Load Balancers to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "loadBalancerName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "loadBalancerArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scheme": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "instances": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "subnetAvailabilityZonePair": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "dnsName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "availabilityZones": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "type": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "createdTime": types.DateTime(),
        "subnets": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "state": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "listeners": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "ipAddressType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "securityGroups": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [prepare_dataclass(load_balancer) for load_balancer in load_balancers]
    )

    # Change the _type column to type:
    df.rename(columns={"_type": "type"}, inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_routetable(
    route_tables: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_routetable_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='routetable') to SQL.

    Args:
        route_tables (BaseList): The BaseList of Route Tables to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "routeTableId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "routes": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "routeDtos": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "associations": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(route_table) for route_table in route_tables])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_ebsvolume(
    ebs_volumes: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_ebsvolume_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='ebsvolume') to SQL.

    Args:
        ebs_volumes (BaseList): The BaseList of EBS Volumes to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "volumeId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "volumeType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "availabilityZone": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "encrypted": types.Boolean(),
        "snapshotId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "state": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "createTime": types.DateTime(),
        "size": types.Integer(),
        "iops": types.Integer(),
        "kmsKeyId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "attachments": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(ebs_volume) for ebs_volume in ebs_volumes])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_autoscalinggroup(
    asgs: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_autoscalinggroup_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='auto scaling group') to SQL.

    Args:
        asgs (BaseList): The BaseList of Auto Scaling Groups to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "autoScalingGroupName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "autoScalingGroupARN": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "healthCheckType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "instances": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "availabilityZones": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "launchConfigurationName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdTime": types.DateTime(),
        "loadBalancerNames": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(asg) for asg in asgs])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_ekscluster(
    eks_clusters: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_ekscluster_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='eks cluster') to SQL.

    Args:
        eks_clusters (BaseList): The BaseList of EKS Clusters to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "associations_resourcesVpcConfig_clusterSecurityGroupId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "associations_resourcesVpcConfig_securityGroupIds": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "associations_resourcesVpcConfig_vpcId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "associations_resourcesVpcConfig_subnetIds": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "version": types.Float(),
        "endpoint": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "resoucesVpcConfig_endpointPrivateAccess": types.Boolean(),
        "resoucesVpcConfig_endpointPublicAccess": types.Boolean(),
        "roleArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "identity_oidc_issuer": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "platformVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "roleName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "logging_clusterLogging": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "arn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(eks_cluster) for eks_cluster in eks_clusters])

    # Drop the associations, resoucesVpcConfig, identity, logging columns:
    df.drop(
        columns=["associations", "resoucesVpcConfig", "identity", "logging"],
        inplace=True,
    )

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_eksnodegroup(
    eks_nodegroups: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_eksnodegroup_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='eks nodegroup') to SQL.

    Args:
        eks_nodegroups (BaseList): The BaseList of EKS Node Groups to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "associations_clusterArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "associations_clusterName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "associations_subnets": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "associations_resources_autoScalingGroups": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "nodeRoleName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scalingConfig_maxSize": types.Integer(),
        "scalingConfig_desiredSize": types.Integer(),
        "scalingConfig_minSize": types.Integer(),
        "capacityType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "instanceTypes": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "releaseVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "version": types.Float(),
        "labels": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "diskSize": types.Integer(),
        "nodeRole": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "launchTemplate_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "launchTemplate_id": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "launchTemplate_version": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "amiType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "health": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [prepare_dataclass(eks_nodegroup) for eks_nodegroup in eks_nodegroups]
    )

    # Drop the associations, scalingConfig, launchTemplate columns:
    df.drop(columns=["associations", "scalingConfig", "launchTemplate"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_fargateprofile(
    fargate_profiles: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_fargateprofile_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='eks fargate profile') to SQL.

    Args:
        fargate_profiles (BaseList): The BaseList of Fargate Profiles to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "associations_clusterArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "associations_clusterName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "associations_subnets": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "podExecutionRoleArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "selectors": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "podExecutionRoleName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [prepare_dataclass(fargate_profile) for fargate_profile in fargate_profiles]
    )

    # Drop the associations column:
    df.drop(columns=["associations"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_vpcendpoint(
    vpc_endpoints: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_vpcendpoint_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='vpc endpoint') to SQL.

    Args:
        vpc_endpoints (BaseList): The BaseList of VPC Endpoints to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "policyDocument": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dnsOption": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "serviceName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "securityGroupSet": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "privateDnsEnabled": types.Boolean(),
        "requesterManaged": types.Boolean(),
        "ipAddressType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "routeTableIds": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "vpcEndpointType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "vpcId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "state": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "dnsEntrySets": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "networkInterfaceIds": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "subnetIds": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(vpc_endpoint) for vpc_endpoint in vpc_endpoints])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_vpcendpointservice(
    vpc_endpoint_services: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_vpcendpointservice_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='vpc endpoint service') to SQL.

    Args:
        vpc_endpoint_services (BaseList): The BaseList of VPC Endpoint Services to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "owner": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "acceptanceRequired": types.Boolean(),
        "managesVpcEndpoints": types.Boolean(),
        "availabilityZone": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "baseEndpointDnsName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "supportedIpAddressType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "serviceTypes": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "privateDnsName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "privateDnsNameVerificationState": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "vpcEndpointPolicySupported": types.Boolean(),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [
            prepare_dataclass(vpc_endpoint_service)
            for vpc_endpoint_service in vpc_endpoint_services
        ]
    )

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_iamgroup(
    iam_groups: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_iamgroup_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='iam group') to SQL.

    Args:
        iam_groups (BaseList): The BaseList of IAM Groups to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "classifications": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "groupId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "GroupPolicyList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "AttachedManagedPolicies": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "arn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(iam_group) for iam_group in iam_groups])

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_iampolicy(
    iam_policies: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_iampolicy_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='iam policy') to SQL.

    Args:
        iam_policies (BaseList): The BaseList of IAM Policies to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "IsAttachable": types.Boolean(),
        "PermissionsBoundaryUsageCount": types.Integer(),
        "AttachmentCount": types.Integer(),
        "DefaultVersionId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultPolicyVersion_VersionId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultPolicyVersion_IsDefaultVersion": types.Boolean(),
        "defaultPolicyVersion_Document_Version": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultPolicyVersion_Document_Statement": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "defaultPolicyVersion_CreateDate": types.DateTime(),
        "type": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "path": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "classifications": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "policyId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "arn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(iam_policy) for iam_policy in iam_policies])

    # Change the _type column to type:
    df.rename(columns={"_type": "type"}, inplace=True)

    # Drop the defaultPolicyVersion column:
    df.drop(columns=["defaultPolicyVersion"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_iamrole(
    iam_roles: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_iamrole_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='iam role') to SQL.

    Args:
        iam_roles (BaseList): The BaseList of IAM Roles to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "AttachedManagedPolicies": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "RoleLastUsed": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "arn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "roleId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "PermissionsBoundary_PermissionsBoundaryArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "PermissionsBoundary_PermissionsBoundaryType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "path": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "classifications": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "AssumeRolePolicyDocument_Version": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "AssumeRolePolicyDocument_Statement": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "trustedEntities": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "RolePolicyList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "InstanceProfileList": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(iam_role) for iam_role in iam_roles])

    # Drop the PermissionsBoundary, AssumeRolePolicyDocument columns:
    df.drop(columns=["PermissionsBoundary", "AssumeRolePolicyDocument"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_sagemakernotebook(
    sagemaker_notebooks: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_sagemakernotebook_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='sagemaker notebook') to SQL.

    Args:
        sagemaker_notebooks (BaseList): The BaseList of SageMaker Notebooks to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "subnetId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "networkInterfaceId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "volumeSizeInGB": types.Float(),
        "rootAccess": types.Boolean(),
        "instanceType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "platformId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "url": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "directInternetAccess": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "roleArn": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "securityGroups": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "defaultCodeRepository": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "kmsKeyId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [
            prepare_dataclass(sagemaker_notebook)
            for sagemaker_notebook in sagemaker_notebooks
        ]
    )

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_totalcloud_aws_cloudfrontdistribution(
    cloudfront_distributions: BaseList,
    cnxn: Connection,
    table_name: str = "totalcloud_aws_cloudfrontdistribution_inventory",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from totalcloud.get_inventory(provider='aws', resource_type='cloudfront distribution') to SQL.

    Args:
        cloudfront_distributions (BaseList): The BaseList of CloudFront Distributions to upload.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to 'vmdr_reports'.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "displayName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "connectorId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scanId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "defaultCacheBehavior_Compress": types.Boolean(),
        "defaultCacheBehavior_FunctionAssociations_Quantity": types.Integer(),
        "defaultCacheBehavior_FunctionAssociations_Items": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "defaultCacheBehavior_LambdaFunctionAssociations_Quantity": types.Integer(),
        "defaultCacheBehavior_LambdaFunctionAssociations_Items": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "defaultCacheBehavior_TargetOriginId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultCacheBehavior_ViewerProtocolPolicy": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultCacheBehavior_TrustedSigners_Enabled": types.Boolean(),
        "defaultCacheBehavior_TrustedSigners_Quantity": types.Integer(),
        "defaultCacheBehavior_FieldLevelEncryptionId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "defaultCacheBehavior_DefaultTTL": types.Integer(),
        "defaultCacheBehavior_TrustedKeyGroups_Enabled": types.Boolean(),
        "defaultCacheBehavior_TrustedKeyGroups_Quantity": types.Integer(),
        "defaultCacheBehavior_AllowedMethods_CachedMethods_Quantity": types.Integer(),
        "defaultCacheBehavior_AllowedMethods_CachedMethods_Items": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "defaultCacheBehavior_AllowedMethods_Items_Method": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "defaultCacheBehavior_SmoothStreaming": types.Boolean(),
        "defaultCacheBehavior_MinTTL": types.Integer(),
        "defaultCacheBehavior_MaxTTL": types.Integer(),
        "loggingEnabled": types.Boolean(),
        "isIpv6Enabled": types.Boolean(),
        "includeCookies": types.Boolean(),
        "geoRestriction_RestrictionType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "geoRestriction_Quantity": types.Integer(),
        "enabled": types.Boolean(),
        "webAclId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "minimumProtocolVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "httpVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "priceClass": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "loggingBucket": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "origins": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "comment": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "id": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "arn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "staging": types.Boolean(),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "loggingBucketPrefix": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "acmCertificateARN": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Add in the base fields:
    COLS.update(BASE_AWS_FIELDS)

    # Convert the BaseList to a DataFrame:
    df = DataFrame(
        [
            prepare_dataclass(cloudfront_distribution)
            for cloudfront_distribution in cloudfront_distributions
        ]
    )

    # Drop the defaultCacheBehavior, geoRestriction columns:
    df.drop(columns=["defaultCacheBehavior", "geoRestriction"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        table_name,
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )



"""
Contains the functions to upload supported Container Security API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList


def upload_was_webapps(
    webapps: BaseList,
    cnxn: Connection,
    table_name: str = "was_webapps",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```was.get_webapps```
    to a SQL database.

    Args:
        webapps (BaseList): A BaseList of WebApp objects.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.Integer(),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "url": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "riskScore": types.Integer(),
        "owner": types.Integer(),
        "tags": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "createdDate": types.DateTime(),
        "updatedDate": types.DateTime(),
    }

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(webapp) for webapp in webapps])

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)



"""
Contains the datatype filter mappings for
operators
"""

ENDPOINT_MAPPINGS = {
    "count_webapps": {
        "id": "INTEGER",
        "name": "TEXT",
        "url": "TEXT",
        "tags_name": "TEXT",
        "tags_id": "INTEGER",
        "createdDate": "DATE",
        "updatedDate": "DATE",
        "isScheduled": "BOOLEAN",
        "isScanned": "BOOLEAN",
        "lastScan_status": "KEYWORD",
        "lastScan_date": "DATE",
    },
    "get_webapps": {
        "id": "INTEGER",
        "name": "TEXT",
        "url": "TEXT",
        "tags_name": "TEXT",
        "tags_id": "INTEGER",
        "createdDate": "DATE",
        "updatedDate": "DATE",
        "isScheduled": "BOOLEAN",
        "isScanned": "BOOLEAN",
        "lastScan_status": "KEYWORD",
        "lastScan_date": "DATE",
        "verbose": "BOOLEAN",
    },
}

FILTER_MAPPING = {
    "INTEGER": ["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"],
    "TEXT": ["CONTAINS", "EQUALS", "NOT EQUALS"],
    "DATE": ["EQUALS", "NOT EQUALS", "GREATER", "LESSER"],
    "KEYWORD": ["EQUALS", "NOT EQUALS", "IN"],
    "BOOLEAN": ["EQUALS", "NOT EQUALS"],
}



"""
hosts.py - contains the dataclass for a Qualys GAV host record.
"""

from dataclasses import dataclass, asdict
from datetime import datetime
from typing import List, Optional, Union

from frozendict import frozendict

from ..base.base_list import BaseList

SOFTWARE_SCHEMA = frozendict(
    {
        "software": {
            "publisher": "",
            "productName": "",
            "version": "",
            "softwareType": "",
            "isIgnored": False,
            "category": "",
            "ignoredReason": "",
            "lastUpdated": "",
            "installPath": "",
            "lifecycle": ["eolDate", "eosDate", "stage", "lifeCycleConfidence"],
            "cpeId": "",
            "cpe": "",
        }
    }
)


def handle_dict_or_list(data: Union[dict, list[dict]]) -> list[dict]:
    """
    Checks if an attribute contains a list of values or a single value,
    putting the single value into a list if necessary.

    Args:
        data (Union[dict, list[dict]]): The data to check.

    Returns:
        list[dict]: A list of dictionaries.
    """

    if isinstance(data, dict):
        data = [data]

    return data


@dataclass
class Host:
    """
    Host - represents a Qualys GAV host record.

    due to the fact that some APIs have excludeFields and IncludeFields parameters,
    virtually all fields are optional other than assetId.
    """

    assetId: int = None
    assetUUID: Optional[str] = None
    hostId: Optional[int] = None
    lastModifiedDate: Optional[str] = None
    agentId: Optional[str] = None
    createdDate: Optional[str] = None
    sensorLastUpdatedDate: Optional[str] = None
    assetType: Optional[str] = None
    address: Optional[str] = None
    dnsName: Optional[str] = None
    assetName: Optional[str] = None
    netbiosName: Optional[str] = None
    timeZone: Optional[str] = None
    biosDescription: Optional[str] = None
    lastBoot: Optional[str] = None
    totalMemory: Optional[int] = None
    cpuCount: Optional[int] = None
    lastLoggedOnUser: Optional[str] = None
    domainRole: Optional[str] = None
    hwUUID: Optional[str] = None
    biosSerialNumber: Optional[str] = None
    biosAssetTag: Optional[str] = None
    isContainerHost: Optional[bool] = None
    operatingSystem: Optional[str] = None
    operatingSystem_osName: Optional[str] = None
    operatingSystem_fullName: Optional[str] = None
    operatingSystem_category: Optional[str] = None
    operatingSystem_category1: Optional[str] = None
    operatingSystem_category2: Optional[str] = None
    operatingSystem_productName: Optional[str] = None
    operatingSystem_publisher: Optional[str] = None
    operatingSystem_edition: Optional[str] = None
    operatingSystem_marketVersion: Optional[str] = None
    operatingSystem_version: Optional[str] = None
    operatingSystem_update: Optional[str] = None
    operatingSystem_architecture: Optional[str] = None
    operatingSystem_lifecycle: Optional[None] = None
    # OperatingSystem_lifecycle is parsed into the following fields:
    operatingSystem_lifecycle_gaDate: Optional[datetime] = None
    operatingSystem_lifecycle_eolDate: Optional[datetime] = None
    operatingSystem_lifecycle_eosDate: Optional[datetime] = None
    operatingSystem_lifecycle_stage: Optional[str] = None
    operatingSystem_lifecycle_lifeCycleConfidence: Optional[str] = None
    operatingSystem_lifecycle_eolSupportStage: Optional[str] = None
    operatingSystem_lifecycle_eosSupportStage: Optional[str] = None
    operatingSystem_lifecycle_detectionScore: Optional[int] = None
    # End of lifecycle fields
    operatingSystem_productUrl: Optional[str] = None
    operatingSystem_productFamily: Optional[str] = None
    operatingSystem_release: Optional[str] = None
    operatingSystem_cpeId: Optional[str] = None
    operatingSystem_cpe: Optional[str] = None
    operatingSystem_cpeType: Optional[str] = None
    operatingSystem_installDate: Optional[datetime] = None
    hardwareVendor: Optional[str] = None
    hardware: Optional[None] = None
    # Hardware is parsed into the following fields:
    hardware_lifecycle_introDate: Optional[datetime] = None
    hardware_lifecycle_gaDate: Optional[datetime] = None
    hardware_lifecycle_eosDate: Optional[datetime] = None
    hardware_lifecycle_obsoleteDate: Optional[datetime] = None
    hardware_lifecycle_stage: Optional[str] = None
    hardware_lifecycle_lifeCycleConfidence: Optional[str] = None
    hardware_fullName: Optional[str] = None
    hardware_category: Optional[str] = None
    hardware_category1: Optional[str] = None
    hardware_category2: Optional[str] = None
    hardware_manufacturer: Optional[str] = None
    hardware_productName: Optional[str] = None
    hardware_model: Optional[str] = None
    hardware_lifecycle: Optional[None] = None
    hardware_productUrl: Optional[str] = None
    hardware_productFamily: Optional[str] = None
    # End of hardware fields
    userAccountListData: Optional[Union[list[dict], BaseList[str]]] = None
    openPortListData: Optional[Union[list[dict], BaseList[str]]] = None
    volumeListData: Optional[Union[list[dict], BaseList[str]]] = None
    networkInterfaceListData: Optional[dict] = None
    softwareListData: Optional[Union[list[dict], BaseList[dict]]] = None
    softwareComponent: Optional[str] = None
    provider: Optional[str] = None
    cloudProvider: Optional[dict] = None
    # AWS:
    cloudProvider_accountId: Optional[str] = None
    cloudProvider_availabilityZone: Optional[str] = None
    cloudProvider_hostname: Optional[str] = None
    cloudProvider_imageId: Optional[str] = None
    cloudProvider_instanceId: Optional[str] = None
    cloudProvider_instanceState: Optional[str] = None
    cloudProvider_instanceType: Optional[str] = None
    cloudProvider_qualysScanner: Optional[bool] = None
    cloudProvider_launchdate: Optional[Union[str, datetime]] = None
    cloudProvider_privateDNS: Optional[str] = None
    cloudProvider_privateIpAddress: Optional[str] = None
    cloudProvider_publicDNS: Optional[str] = None
    cloudProvider_publicIpAddress: Optional[str] = None
    cloudProvider_hasAgent: Optional[bool] = None
    cloudProvider_region: Optional[str] = None
    cloudProvider_spotInstance: Optional[bool] = None
    cloudProvider_subnetId: Optional[str] = None
    cloudProvider_vpcId: Optional[str] = None
    # Azure:
    cloudProvider_imageOffer: Optional[str] = None
    cloudProvider_imagePublisher: Optional[str] = None
    cloudProvider_imageVersion: Optional[str] = None
    cloudProvider_location: Optional[str] = None
    cloudProvider_macAddress: Optional[str] = None
    cloudProvider_name: Optional[str] = None
    cloudProvider_platform: Optional[str] = None
    cloudProvider_resourceGroupName: Optional[str] = None
    cloudProvider_size: Optional[str] = None
    cloudProvider_state: Optional[str] = None
    cloudProvider_subnet: Optional[str] = None
    cloudProvider_subscriptionId: Optional[str] = None
    cloudProvider_virtualNetwork: Optional[str] = None
    cloudProvider_vmId: Optional[str] = None
    # NO SUPPORT FOR OTHER CLOUD PROVIDERS YET!
    agent: Optional[dict] = None
    agent_version: Optional[str] = None
    agent_configurationProfile: Optional[str] = None
    agent_connectedFrom: Optional[str] = None
    agent_lastActivity: Optional[Union[str, datetime]] = None
    agent_lastCheckedIn: Optional[Union[str, datetime]] = None
    agent_lastInventory: Optional[Union[str, datetime]] = None
    agent_udcManifestAssigned: Optional[bool] = None
    agent_errorStatus: Optional[bool] = None
    agent_key: Optional[str] = None
    agent_status: Optional[str] = None
    sensor: Optional[dict] = None
    sensor_activatedForModules: Optional[Union[list[str], BaseList[str]]] = None
    sensor_pendingActivationForModules: Optional[Union[list[str], BaseList[str]]] = None
    sensor_lastVMScan: Optional[Union[str, datetime]] = None
    sensor_lastComplianceScan: Optional[Union[str, datetime]] = None
    sensor_lastFullScan: Optional[Union[str, datetime]] = None
    sensor_lastVmScanDateScanner: Optional[Union[str, datetime]] = None
    sensor_lastVmScanDateAgent: Optional[Union[str, datetime]] = None
    sensor_lastPcScanDateScanner: Optional[Union[str, datetime]] = None
    sensor_lastPcScanDateAgent: Optional[Union[str, datetime]] = None
    sensor_firstEasmScanDate: Optional[Union[str, datetime]] = None
    sensor_lastEasmScanDate: Optional[Union[str, datetime]] = None
    container: Optional[dict] = None
    container_product: Optional[str] = None
    container_version: Optional[str] = None
    container_noOfContainers: Optional[int] = None
    container_noOfImages: Optional[int] = None
    container_hasSensor: Optional[bool] = None
    inventory: Optional[dict] = None
    inventory_source: Optional[str] = None
    inventory_created: Optional[Union[str, datetime]] = None
    inventory_lastUpdated: Optional[Union[str, datetime]] = None
    activity: Optional[dict] = None
    activity_source: Optional[str] = None
    activity_lastScannedDate: Optional[Union[str, datetime]] = None
    tagList: Optional[Union[str, List[str], BaseList[str]]] = None
    serviceList: Optional[Union[str, List[str], BaseList[str]]] = None
    lastLocation: Optional[Union[dict, str]] = None
    criticality: Optional[int] = None
    businessInformation: Optional[dict] = None
    assignedLocation: Optional[dict] = None
    businessAppListData: Optional[Union[list[dict], BaseList[str]]] = None
    riskScore: Optional[int] = None
    passiveSensor: Optional[dict] = None
    domain: Optional[str] = None
    subdomain: Optional[str] = None
    missingSoftware: Optional[Union[list[dict], BaseList[str]]] = None
    whois: Optional[dict] = None
    organizationName: Optional[str] = None
    isp: Optional[str] = None
    asn: Optional[str] = None
    easmTags: Optional[Union[str, List[str], BaseList[str]]] = None
    hostingCategory1: Optional[str] = None
    customAttributes: Optional[dict] = None
    processor: Optional[Union[dict, str]] = None

    def __post_init__(self):
        DT_FIELDS = [
            "lastModifiedDate",
            "createdDate",
            "sensorLastUpdatedDate",
            "lastBoot",
        ]

        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        if self.businessAppListData:
            # EXPERIMENTAL! I DO NOT HAVE ANY BUSINESS APPS!
            try:
                data = handle_dict_or_list(self.businessAppListData["businessApp"])
                bl = BaseList()
                bl.extend([app.get("name") for app in data])
                setattr(self, "businessAppListData", bl)
            except KeyError:
                # Best guesses here...
                data = handle_dict_or_list(self.businessAppListData["app"])
                bl = BaseList()
                bl.extend([app.get("name") for app in data])
                setattr(self, "businessAppListData", bl)

        if self.operatingSystem:
            for field in [
                "osName",
                "fullName",
                "category",
                "category1",
                "category2",
                "productName",
                "publisher",
                "edition",
                "marketVersion",
                "version",
                "update",
                "architecture",
                "lifecycle",
                "productUrl",
                "productFamily",
                "release",
                "cpeId",
                "cpe",
                "cpeType",
            ]:
                if self.operatingSystem.get(field):
                    if self.operatingSystem.get(field) in [",", ",,", "Not Announced"]:
                        setattr(self, f"operatingSystem_{field}", None)
                    else:
                        setattr(
                            self,
                            f"operatingSystem_{field}",
                            self.operatingSystem[field],
                        )
            # we will deal with installDate separately:
            if self.operatingSystem.get("installDate"):
                # Convert installDate to datetime object
                setattr(
                    self,
                    "operatingSystem_installDate",
                    datetime.fromisoformat(self.operatingSystem["installDate"]),
                )

            # Set the operatingSystem field to None
            setattr(self, "operatingSystem", None)

        if self.hardware:
            for field in [
                "fullName",
                "category",
                "category1",
                "category2",
                "manufacturer",
                "productName",
                "model",
                "productUrl",
                "productFamily",
            ]:
                if self.hardware.get(field):
                    if self.hardware.get(field) in [",", ",,", "Not Announced"]:
                        setattr(self, f"hardware_{field}", None)
                    else:
                        setattr(self, f"hardware_{field}", self.hardware[field])

            # for field in ["introDate", "gaDate", "eosDate", "obsoleteDate"]:
            #    if self.hardware.get(field):
            #        setattr(
            #            self,
            #            f"hardware_{field}",
            #            datetime.fromisoformat(self.hardware[field]),
            #        )

            if self.hardware.get("lifecycle"):
                for field in ["introDate", "gaDate", "eosDate", "obsoleteDate"]:
                    if self.hardware["lifecycle"].get(field):
                        if self.hardware["lifecycle"].get(field) in [
                            ",",
                            ",,",
                            "Not Announced",
                        ]:
                            setattr(self, f"hardware_lifecycle_{field}", None)

                        else:
                            setattr(
                                self,
                                f"hardware_lifecycle_{field}",
                                datetime.fromisoformat(
                                    self.hardware["lifecycle"][field]
                                ),
                            )
                for field in ["stage", "lifeCycleConfidence"]:
                    if self.hardware["lifecycle"].get(field):
                        setattr(
                            self,
                            f"hardware_lifecycle_{field}",
                            self.hardware["lifecycle"][field],
                        )

            # Set the hardware field to None
            setattr(self, "hardware", None)

        if self.userAccountListData:
            # Check for a dict or a list of dicts:
            data = handle_dict_or_list(self.userAccountListData["userAccount"])
            bl = BaseList()
            for user in data:
                bl.append(user.get("name"))
            setattr(self, "userAccountListData", bl)

        if self.openPortListData:
            # Check for a dict or a list of dicts:
            data = self.openPortListData["openPort"]
            bl = BaseList()
            for port in data:
                bl.append(
                    f"{port.get('port')}-{port.get('protocol')} ({port.get('detectedService')})"
                )
            setattr(self, "openPortListData", bl)

        if self.volumeListData:
            # Check for a dict or a list of dicts:
            data = handle_dict_or_list(self.volumeListData["volume"])
            bl = BaseList()
            for vol in data:
                try:
                    percent_filled = (
                        (vol.get("size") - vol.get("free")) / vol.get("size") * 100
                    )
                except ZeroDivisionError:
                    percent_filled = 0.0
                bl.append(f"{vol.get('name')}: {percent_filled:.2f}% filled")
            setattr(self, "volumeListData", bl)

        if self.networkInterfaceListData:
            # Check for a dict or a list of dicts:
            data = handle_dict_or_list(
                self.networkInterfaceListData["networkInterface"]
            )
            bl = BaseList()
            for iface in data:
                bl.append(
                    f"{iface.get('interfaceName').replace('      ', ' ')} - {iface.get('manufacturer')}"
                )  # Replace multi-spaces with single spaces for easier reading
            setattr(self, "networkInterfaceListData", bl)

        if self.softwareListData:
            data = handle_dict_or_list(self.softwareListData["software"])
            bl = BaseList()
            for sw in data:
                #                bl.append(
                #                    f"{sw.get('fullName')} ({sw.get('category')}) ({sw.get('ignoredReason')})"
                #                )
                sw_info = {}
                for k, v in SOFTWARE_SCHEMA["software"].items():
                    # If the key doesn't exist, don't add.
                    # This helps with SQL inserts by not adding
                    # null values.
                    if sw.get(k):
                        if isinstance(v, list):
                            for sub_k in v:
                                if sw[k].get(sub_k):
                                    sw_info[sub_k] = sw[k][sub_k]
                        else:
                            sw_info[k] = sw[k]
                bl.append(sw_info)
            setattr(self, "softwareListData", bl)

        if self.softwareComponent:
            if not isinstance(self.softwareComponent, str):
                raise Exception("SoftwareComponent must be a string.")

        if self.cloudProvider:
            # A bit different. This is a dictionary with all cloud providers.
            # The valid one will have a dictionary underneath of it.
            # First, find the one that is not a NoneType:
            for provider, data in self.cloudProvider.items():
                if data:
                    cloudProvider = provider
                    subkeys = list(data.keys())
                    break

            for subkey in subkeys:
                if subkey != "tags":
                    for attr in [
                        "accountId",
                        "availabilityZone",
                        "hasAgent",
                        "hostname",
                        "imageId",
                        "instanceId",
                        "instanceState",
                        "instanceType",
                        "qualyScanner",
                        "kernelId",
                        "privateDNS",
                        "privateIpAddress",
                        "publicDNS",
                        "publicIpAddress",
                        "spotInstance",
                        "subnetId",
                        "vpcId",
                        "imageOffer",
                        "imagePublisher",
                        "imageVersion",
                        "location",
                        "macAddress",
                        "name",
                        "platform",
                        "resourceGroupName",
                        "size",
                        "state",
                        "subnet",
                        "subscriptionId",
                        "virtualNetwork",
                        "vmId",
                    ]:
                        if self.cloudProvider[cloudProvider].get(subkey).get(attr):
                            setattr(
                                self,
                                f"cloudProvider_{attr}",
                                self.cloudProvider[cloudProvider][subkey].get(attr),
                            )

                    # Convert launchdate to datetime object
                    if self.cloudProvider[cloudProvider].get(subkey).get(
                        "launchDate"
                    ) and not isinstance(
                        self.cloudProvider[cloudProvider][subkey]["launchDate"],
                        datetime,
                    ):
                        setattr(
                            self,
                            "cloudProvider_launchDate",
                            datetime.fromisoformat(
                                self.cloudProvider[cloudProvider][subkey]["launchDate"]
                            ),
                        )

                    # Parse out region:
                    if self.cloudProvider[cloudProvider].get(subkey).get(
                        "region"
                    ) and not isinstance(
                        self.cloudProvider[cloudProvider][subkey]["region"], str
                    ):
                        setattr(
                            self,
                            "cloudProvider_region",
                            self.cloudProvider[cloudProvider][subkey]
                            .get("region")
                            .get("code"),
                        )

                elif subkey == "tags":
                    data = handle_dict_or_list(
                        self.cloudProvider[cloudProvider][subkey]
                    )
                    bl = BaseList()
                    if data:
                        for tag in data:
                            s = (
                                f"{tag.get('key')}:{tag.get('value')}"
                                if tag.get("key")
                                else f"{tag.get('name')}:{tag.get('value')}"
                            )
                        setattr(self, "cloudProvider_tags", bl)

                else:
                    print(f"Unknown subkey: {subkey}")

                # Set the cloudProvider field to the valid provider
            setattr(self, "cloudProvider", cloudProvider)

        if self.agent:
            for field in [
                "version",
                "configurationProfile",
                "connectedFrom",
                "udcManifestAssigned",
                "errorStatus",
            ]:
                if self.agent.get(field):
                    setattr(self, f"agent_{field}", self.agent[field])
            if self.agent.get("activations"):
                setattr(self, "agent_key", self.agent.get("activations")[0].get("key"))
                setattr(
                    self, "agent_status", self.agent.get("activations")[0].get("status")
                )
            for dt_field in ["lastActivity", "lastCheckedIn", "lastInventory"]:
                if self.agent.get(dt_field) and not isinstance(
                    self.agent.get(dt_field), datetime
                ):
                    if self.agent.get(dt_field) != -1:
                        setattr(
                            self,
                            f"agent_{dt_field}",
                            datetime.fromtimestamp(self.agent[dt_field] / 1000),
                        )
                    else:
                        setattr(self, f"agent_{dt_field}", None)

            # Set the agent field to None
            setattr(self, "agent", None)

        if self.sensor:
            for field in ["activatedForModules", "pendingActivationForModules"]:
                bl = BaseList()
                if self.sensor.get(field):
                    bl.extend(self.sensor[field])
                    setattr(self, f"sensor_{field}", bl)
            for dt_field in [
                "lastVMScan",
                "lastComplianceScan",
                "lastFullScan",
                "lastVmScanDateScanner",
                "lastVmScanDateAgent",
                "lastPcScanDateScanner",
                "lastPcScanDateAgent",
                "firstEasmScanDate",
                "lastEasmScanDate",
            ]:
                if self.sensor.get(dt_field) and not isinstance(
                    self.sensor.get(dt_field), datetime
                ):
                    setattr(
                        self,
                        f"sensor_{dt_field}",
                        datetime.fromtimestamp(self.sensor[dt_field] / 1000),
                    )

            # Set the sensor field to None
            setattr(self, "sensor", None)

        if self.container:
            for field in ["product", "version", "noOfContainers", "noOfImages"]:
                if self.container.get(field):
                    setattr(self, f"container_{field}", self.container[field])

            if self.container.get("hasSensor"):
                setattr(self, "container_hasSensor", bool(self.container["hasSensor"]))
            else:
                setattr(self, "container_hasSensor", False)

            # Set the container field to None
            setattr(self, "container", None)

        if self.inventory:
            setattr(self, "inventory_source", self.inventory.get("source"))
            setattr(
                self,
                "inventory_created",
                datetime.fromtimestamp(self.inventory.get("created") / 1000),
            )
            setattr(
                self,
                "inventory_lastUpdated",
                datetime.fromtimestamp(self.inventory.get("lastUpdated") / 1000),
            )

            # Set the inventory field to None
            setattr(self, "inventory", None)

        if self.activity:
            setattr(self, "activity_source", self.activity.get("source"))
            setattr(
                self,
                "activity_lastScannedDate",
                datetime.fromtimestamp(self.activity.get("lastScannedDate") / 1000),
            )

            # Set the activity field to None
            setattr(self, "activity", None)

        if self.tagList:
            data = handle_dict_or_list(self.tagList["tag"])
            bl = BaseList()
            bl.extend([tag.get("tagName") for tag in data])
            setattr(self, "tagList", bl)

        if self.serviceList:
            data = handle_dict_or_list(self.serviceList["service"])
            bl = BaseList()
            for service in data:
                bl.append(f"{service.get('name')} ({service.get('status')})")
            setattr(self, "serviceList", bl)

        if self.lastLocation and not isinstance(self.lastLocation, str):
            setattr(self, "lastLocation", self.lastLocation.get("name"))

        if self.criticality:
            score = self.criticality.get("score")
            if not score:
                score = 0
            setattr(self, "criticality", score)

        if self.missingSoftware:
            data = handle_dict_or_list(self.missingSoftware)
            bl = BaseList()
            if data:
                for sw in data:
                    full_category = f"{sw.get('category1')} / {sw.get('category2')}"
                    bl.append(f"{sw.get('name')} ({full_category})")

                setattr(self, "missingSoftware", bl)
            else:
                setattr(self, "missingSoftware", None)

        if self.easmTags:
            data = handle_dict_or_list(self.tagList)
            bl = BaseList()
            bl.extend([tag for tag in data])
            setattr(self, "easmTags", bl)

        if self.customAttributes:
            print(self.customAttributes)

        if self.processor and not isinstance(self.processor, str):
            setattr(self, "processor", self.processor.get("description"))

        if self.operatingSystem_lifecycle:
            for field in [
                "gaDate",
                "eolDate",
                "eosDate",
            ]:
                if self.operatingSystem_lifecycle.get(field):
                    if self.operatingSystem_lifecycle.get(field) != "Not Announced":
                        setattr(
                            self,
                            f"operatingSystem_lifecycle_{field}",
                            datetime.fromisoformat(
                                self.operatingSystem_lifecycle[field]
                            ),
                        )
                    else:
                        setattr(self, f"operatingSystem_lifecycle_{field}", None)

            for field in [
                "stage",
                "lifeCycleConfidence",
                "eolSupportStage",
                "eosSupportStage",
            ]:
                if self.operatingSystem_lifecycle.get(field):
                    if self.operatingSystem_lifecycle.get(field) in [
                        ",",
                        ",,",
                        "Not Announced",
                        " ",
                    ]:
                        setattr(self, f"operatingSystem_lifecycle_{field}", None)
                    setattr(
                        self,
                        f"operatingSystem_lifecycle_{field}",
                        self.operatingSystem_lifecycle[field],
                    )

            if self.operatingSystem_lifecycle.get("detectionScore"):
                setattr(
                    self,
                    "operatingSystem_lifecycle_detectionScore",
                    self.operatingSystem_lifecycle["detectionScore"],
                )

            setattr(self, "operatingSystem_lifecycle", None)

    def is_cloud_host(self) -> bool:
        """
        Returns True if the host is a cloud host, False otherwise.
        """
        return self.cloudProvider is not None

    def is_container_host(self) -> bool:
        """
        Returns True if the host is a container host, False otherwise.
        """
        return self.container is not None

    def has_agent(self) -> bool:
        """
        Returns True if the host has an agent, False otherwise.
        """
        return self.agentId is not None

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the host object.
        """
        return asdict(self)

    def keys(self) -> list:
        """
        Returns a list of keys for the host object.
        """
        return [key for key in self.to_dict().keys()]

    def values(self) -> list:
        """
        Returns a list of values for the host object.
        """
        return [value for value in self.to_dict().values()]

    def items(self) -> list:
        """
        Returns a list of tuples for the host object.
        """
        return [(key, value) for key, value in self.to_dict().items()]

    def valid_values(self) -> list:
        """
        Return a list of keys that have values.
        """
        return [key for key, value in self.to_dict().items() if value]

    def __dict__(self) -> dict:
        return asdict(self)

    def __iter__(self):
        """
        Allows for iteration over the host object.
        """
        for key, value in self.to_dict().items():
            yield key, value

    def __str__(self) -> str:
        """
        String representation of the host object.
        """
        return self.assetName

    def __int__(self) -> int:
        """
        Integer representation of the host object.
        """
        return self.assetId

    def __repr__(self) -> str:
        """
        Breaking the unwritten rule of having repr be something
        that can create the object for terminal space's sake.
        """
        return f"AssetID({self.assetId})"



"""
Contains functions to interact with Web applications in the Qualys WAS module.
"""

from typing import Union

from .data_classes.WebApp import WebApp
from .base.service_requests import build_service_request
from .base.parse_kwargs import validate_kwargs
from .base.service_requests import validate_response
from ..base.call_api import call_api
from ..auth.basic import BasicAuth
from ..exceptions.Exceptions import QualysAPIError
from ..base.base_list import BaseList


def count_webapps(auth: BasicAuth, **kwargs) -> int:
    """
    Return a count of web applications in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - url (str): Web application URL.
        - url_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the URL filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isScheduled (bool): Whether the web application has a scan scheduled.
        - isScheduled_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScheduled filter.
        - isScanned (bool): Whether the web application has been scanned.
        - isScanned_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScanned filter.
        - lastScan_status (Literal["SUBMITTED", "RUNNING", "FINISHED", "TIME_LIMIT_EXCEEDED", "SCAN_NOT_LAUNCHED", "SCANNER_NOT_AVAILABLE", "ERROR", "CANCELLED"]): Status of the last scan.
        - lastScan_status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.

    Returns:
        int: The count of web applications that match the filters.
    """

    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="count_webapps", **kwargs)
        payload = build_service_request(**kwargs)

    # Make the API call
    response = call_api(
        auth=auth,
        module="was",
        endpoint="count_webapps",
        payload=payload,
        headers={"Content-Type": "text/xml"},
    )

    # Parse the XML response
    parsed = validate_response(response)

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        raise QualysAPIError(
            f"API response returned error: {serviceResponse.get('responseCode')}"
        )

    return int(serviceResponse.get("count"))


def get_webapps(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> object:
    """
    Get a list of web applications in the Qualys WAS module
    according to the provided filters.

    Args:
        auth (BasicAuth): The authentication object.
        page_count (int): The number of pages to return. If 'all', return all pages. Default is 'all'.

    ## Kwargs:

        - id (Union[str, int]): Web application ID.
        - id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the ID filter.
        - name (str): Web application name.
        - name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the name filter.
        - url (str): Web application URL.
        - url_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the URL filter.
        - tags_name (str): Tag name.
        - tags_name_operator (Literal["CONTAINS", "EQUALS", "NOT EQUALS"]): Operator for the tag name filter.
        - tags_id (Union[str, int]): Tag ID.
        - tags_id_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"]): Operator for the tag ID filter.
        - createdDate (str): Date the web application was created in UTC date/time format.
        - createdDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the created date filter.
        - updatedDate (str): Date the web application was last updated in UTC date/time format.
        - updatedDate_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the updated date filter.
        - isScheduled (bool): Whether the web application has a scan scheduled.
        - isScheduled_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScheduled filter.
        - isScanned (bool): Whether the web application has been scanned.
        - isScanned_operator (Literal["EQUALS", "NOT EQUALS"]): Operator for the isScanned filter.
        - lastScan_status (Literal["SUBMITTED", "RUNNING", "FINISHED", "TIME_LIMIT_EXCEEDED", "SCAN_NOT_LAUNCHED", "SCANNER_NOT_AVAILABLE", "ERROR", "CANCELLED"]): Status of the last scan.
        - lastScan_status_operator (Literal["EQUALS", "NOT EQUALS", "IN"]): Operator for the last scan status filter.
        - lastScan_date (str): Date of the last scan in UTC date/time format.
        - lastScan_date_operator (Literal["EQUALS", "NOT EQUALS", "GREATER", "LESSER"]): Operator for the last scan date filter.
        - verbose (bool): Whether to return verbose output.

    Returns:
        object: The web applications that match the filters.
    """

    # Ensure that page_count is either the string 'all' or an integer >= 1:
    if isinstance(page_count, int) and page_count < 1:
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")
    elif isinstance(page_count, str) and page_count != "all":
        raise QualysAPIError("page_count must be 'all' or an integer >= 1.")

    pageNo = 0
    payload = None

    # If kwargs are provided, validate them:
    if kwargs:
        kwargs = validate_kwargs(endpoint="get_webapps", **kwargs)
        payload = build_service_request(**kwargs)

    appList = BaseList()

    while True:
        # Make the API call
        response = call_api(
            auth=auth,
            module="was",
            endpoint="get_webapps",
            payload=payload,
            headers={"Content-Type": "text/xml"},
        )

        parsed = validate_response(response)

        # Parse the XML response:
        serviceResponse = parsed.get("ServiceResponse")
        if not serviceResponse:
            raise QualysAPIError("No ServiceResponse tag returned in the API response")

        if serviceResponse.get("responseCode") != "SUCCESS":
            raise QualysAPIError(
                f"API response returned error: {serviceResponse.get('responseCode')}"
            )

        if serviceResponse.get("count") == "0":
            print(f"No web applications found on page {pageNo}. Exiting.")
            break

        data = serviceResponse.get("data")

        if data.get("WebApp"):
            data = data.get("WebApp")

        if isinstance(data, dict):
            data = [data]

        for webapp in data:
            # Create the objects:
            appList.append(WebApp.from_dict(webapp))

        print(
            f"Retrieved {serviceResponse.get('count')} web applications on page {pageNo}. Running total: {len(appList)}"
        )

        pageNo += 1

        if page_count != "all" and pageNo >= page_count:
            print(f"Reached page_count limit. Returning {pageNo} page(s).")
            break

        # Check for pagination:
        if serviceResponse.get("hasMoreRecords") == "true":
            # Update the XML payload with the new Criteria:
            # <Criteria field="id" operator="GREATER">XXX</Criteria>
            kwargs["id.operator"] = "GREATER"
            kwargs["id"] = serviceResponse.get("lastId")
            payload = build_service_request(**kwargs)
        else:
            break

    return appList



"""
threat_intel.py - contains the ThreatIntel dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *


@dataclass
class ThreatIntel:
    """
    ThreatIntel - represents a single threat intel entry in a ThreatIntelList.
    """

    ID: int = field(metadata={"description": "The ID of the threat intel."})
    TEXT: str = field(
        metadata={"description": "The text of the threat intel."},
        default="",
        compare=False,
    )

    def __post_init__(self):
        # make sure that the ID is an integer:
        if not isinstance(self.ID, int):
            self.ID = int(self.ID)
        # and that text is a string:
        if not isinstance(self.TEXT, str):
            self.TEXT = str(self.TEXT)

    def __str__(self):
        return str(self.TEXT)

    def __contains__(self, item):
        return item in self.ID or item in self.TEXT

    def copy(self):
        return ThreatIntel(ID=self.ID, TEXT=self.TEXT)

    def is_id(self, id: int):
        return self.ID == id

    def is_text(self, text: str):
        return self.TEXT == text

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a ThreatIntel object from a dictionary.

        Params:
            data (dict): The dictionary containing the data for the ThreatIntel object.

        Returns:
            ThreatIntel: The ThreatIntel object created from the dictionary.
        """
        required_keys = {"ID"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )
        return cls(**data)



"""
query_assets.py - contains the query_assets function for the Global AssetView API (GAV) module.

Gets all assets that satisfy a Qualys Query Language (QQL) filter.
"""

from typing import Union

from ..base.base_list import BaseList
from ..base.call_api import call_api
from ..auth.token import TokenAuth
from ..exceptions.Exceptions import *
from .hosts import Host


def query_assets(
    auth: TokenAuth, page_count: Union["all", int] = "all", **kwargs
) -> BaseList[Host]:
    """
    Queries GAV inventory for assets that satisfy a Qualys Query Language (QQL) filter.

    Params:
        auth (TokenAuth): The authentication object.
        page_count (int): The number of pages to get. Defaults to 'all'.

    ## Kwargs:

        filter (str): The Qualys QQL filter to use.
        excludeFields (str): The fields to exclude.
        includeFields (str): The fields to include.
        lastSeenAssetId (int): The last seen asset ID. Used for automatic pagination.
        lastModifiedDate (str): The last modified date.
        pageSize (int): The number of assets to get per page.

    Returns:
        BaseList[Host]: BaseList of Host objects.
    """

    responses = BaseList()
    pulled = 0

    while True:
        # make the request:
        response = call_api(
            auth=auth, module="gav", endpoint="query_assets", params=kwargs
        )
        # if there is no response, break the loop
        if not response.text:
            print("No Results returned.")
            break

        j = response.json()
        for record in j["assetListData"]["asset"]:
            responses.append(Host(**record))
        (
            print(f"Page {pulled+1} of {page_count} complete.")
            if page_count != "all"
            else print(f"Page {pulled+1} complete.")
        )
        pulled += 1

        if not j["hasMore"]:
            print("No more records.")
            break

        if page_count != "all" and pulled >= page_count:
            print("Page count reached.")
            break

        else:
            kwargs["lastSeenAssetId"] = j["lastSeenAssetId"]

    print("All pages complete.")
    return responses



"""
Code to generate XML service requests to Qualys WAS API
"""

from requests import Response

from ...base.xml_parser import xml_parser
from ...exceptions.Exceptions import QualysAPIError


def build_service_request(**kwargs):
    """
    Build the XML request. Users pass in the filters as kwargs.
    there is the filter itself as well as a filter_operator kwarg.
    If the filter_operator is not provided, default to "EQUALS".
    Do not create a Criteria tag for any _operator kwargs.
    """

    xml = "<ServiceRequest>{PREFERENCES}<filters>{REPLACE}</filters></ServiceRequest>"
    user_xml = ""

    # Check for any tags that need to be under preferences instead of filters:
    PREFERENCE_TAGS = ["verbose"]

    for tag in PREFERENCE_TAGS:
        if kwargs.get(tag):
            # If boolean, convert to string:
            if isinstance(kwargs.get(tag), bool):
                kwargs[tag] = str(kwargs.get(tag)).lower()
            # If this is the first preference found,
            # build the structure. If not, just add the line of
            # XML to the existing structure.
            if "{PREFERENCES}" in xml:
                xml = xml.replace(
                    "{PREFERENCES}",
                    f"<preferences><{tag}>{kwargs.pop(tag)}</{tag}></preferences>",
                )
            else:
                # Add the key as a tag within the preferences tag.
                xml = xml.replace(
                    "</preferences>", f"<{tag}>{kwargs.pop(tag)}</{tag}</preferences>"
                )

    # If no preferences were found, remove the placeholder:
    xml = xml.replace("{PREFERENCES}", "")

    for kwarg, value in kwargs.items():
        if kwarg.endswith("operator"):
            continue
        # If boolean, convert to string:
        if isinstance(value, bool):
            value = str(value).lower()

        user_xml += f'<Criteria field="{kwarg}" operator="{kwargs.get(kwarg + ".operator", "EQUALS")}">{value}</Criteria>'

    xml = xml.replace("{REPLACE}", user_xml)
    payload = {"_xml_data": xml}

    # if no filters were provided, remove the filters tag:
    if not user_xml:
        payload = {"_xml_data": xml.replace("<filters></filters>", "")}

    return payload


def validate_response(response: Response) -> dict:
    """
    Parse the XML response from the Qualys API
    and check for errors.

    Args:
        response (Response): The response object from the API call.

    Returns:
        dict: The parsed XML response.
    """
    parsed = xml_parser(response.text)

    if response.status_code != 200:
        errorMessage = (
            parsed.get("ServiceResponse")
            .get("responseErrorDetails")
            .get("errorMessage")
        )
        responseCode = parsed.get("ServiceResponse").get("responseCode")
        raise QualysAPIError(
            f"Error retrieving web applications. Status code: {response.status_code}. Endpoint reporting: {responseCode}. Error message: {errorMessage}"
        )

    serviceResponse = parsed.get("ServiceResponse")
    if not serviceResponse:
        raise QualysAPIError("No ServiceResponse tag returned in the API response")

    if serviceResponse.get("responseCode") != "SUCCESS":
        errorMessage = (
            parsed.get("ServiceResponse")
            .get("responseErrorDetails")
            .get("errorMessage")
        )
        responseCode = parsed.get("ServiceResponse").get("responseCode")
        raise QualysAPIError(
            f"Error retrieving web applications. Status code: {response.status_code}. Endpoint reporting: {responseCode}. Error message: {errorMessage}"
        )

    return parsed



"""
CloudView data_classes module

This module contains ways to interact with the Qualys CloudView APIs. 
"""

from .Connectors import AWSConnector, AzureConnector
from .Controls import Control, AccountLevelControl
from .Evaluation import Evaluation, AccountLevelEvaluation
from .AWSResources import AWSBucket, AWSNetworkACL, AWSRDS



"""
Contains the WebApp class for WAS
"""

from datetime import datetime
from typing import Union
from dataclasses import dataclass, asdict

from .Tag import WASTag
from ...base.base_list import BaseList


@dataclass
class WebApp:
    """
    Represents a single Web Application in WAS
    """

    id: int = None
    name: str = None
    url: str = None
    riskScore: int = None
    owner: Union[dict, int] = None
    tags: Union[list, BaseList] = None
    createdDate: Union[str, datetime] = None
    updatedDate: Union[str, datetime] = None

    def __post_init__(self):
        DT_FIELDS = ["createdDate", "updatedDate"]
        INT_FIELDS = ["id", "riskScore"]

        for field in DT_FIELDS:
            value = getattr(self, field)
            if not isinstance(value, datetime):
                setattr(self, field, datetime.fromisoformat(value))

        for field in INT_FIELDS:
            value = getattr(self, field)
            if value and not isinstance(value, int):
                setattr(self, field, int(value))

        if self.owner:
            data = int(self.owner.get("id"))
            setattr(self, "owner", data)

        if "list" in self.tags.keys():
            bl = BaseList()
            data = self.tags.get("list").get("Tag")
            if isinstance(data, dict):
                data = [data]
            for tag in data:
                bl.append(WASTag.from_dict(tag))
            setattr(self, "tags", bl)
        else:
            setattr(self, "tags", None)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    def risk_rating(self) -> str:
        """
        Categorized risk rating based on riskScore.
        0-250: Low
        251-500: Medium
        501-750: High
        751-1000: Critical

        Returns:
            str: The risk rating.
        """

        if not type(self.riskScore) in [int, float]:
            raise ValueError(
                f"riskScore must be an integer or float, not {type(self.riskScore)}"
            )

        if self.riskScore < 251:
            return "Low"
        elif self.riskScore < 501:
            return "Medium"
        elif self.riskScore < 751:
            return "High"
        elif self.riskScore > 750:
            return "Critical"
        # Not needed, but added for clarity:
        else:
            return "Unknown"

    @staticmethod
    def from_dict(data):
        return WebApp(**data)
